class DynamodbStreamRouterException(Exception):
    pass


class KeywordError(DynamodbStreamRouterException):
    pass


class RouteAlreadyExistsException(DynamodbStreamRouterException):
    pass


class SyntaxError(DynamodbStreamRouterException):
    pass



# pyright: reportUndefinedVariable=false
import re
from decimal import Decimal

import simplejson as json
from boto3.dynamodb.types import Binary
from sly import Parser
from sly.yacc import YaccProduction

from ..exceptions import KeywordError, SyntaxError
from . import Condition
from .lexer import ExpressionLexer


class ExpressionParser(Parser):

    _expression_cache = {}

    # Get the token list from the lexer (required)
    tokens = ExpressionLexer.tokens

    # Define precendence
    precedence = (
        ("left", OR),  # noqa: 821
        ("left", AND),  # noqa: 821
        ("right", NOT),  # noqa: 821
        ("nonassoc", EQ, NE, GT, LT, LTE, GTE),  # noqa: 821
    )

    @staticmethod
    def __strip_quotes(val: str) -> str:
        return val[1:-1]

    def parse(self, expression: str) -> Condition:
        if expression not in self._expression_cache:
            self._expression_cache[expression] = super().parse(
                ExpressionLexer().tokenize(expression)
            )
        return self._expression_cache[expression]

    # Grammar rules and actions
    @_("operand EQ operand")  # noqa: 821
    def condition(self, p):
        operand0 = p.operand0
        operand1 = p.operand1
        return lambda m: operand0(m) == operand1(m)

    @_("operand NE operand")  # noqa: 821
    def condition(self, p):  # noqa: 811
        operand0 = p.operand0
        operand1 = p.operand1
        return lambda m: operand0(m) != operand1(m)

    @_("operand GT operand")  # noqa: 821
    def condition(self, p):  # noqa: 811
        operand0 = p.operand0
        operand1 = p.operand1
        return lambda m: operand0(m) > operand1(m)

    @_("operand GTE operand")  # noqa: 821
    def condition(self, p):  # noqa: 811
        operand0 = p.operand0
        operand1 = p.operand1
        return lambda m: operand0(m) >= operand1(m)

    @_("operand LT operand")  # noqa: 821
    def condition(self, p):  # noqa: 811
        operand0 = p.operand0
        operand1 = p.operand1
        return lambda m: operand0(m) < operand1(m)

    @_("operand LTE operand")  # noqa: 821
    def condition(self, p):  # noqa: 811
        operand0 = p.operand0
        operand1 = p.operand1
        return lambda m: operand0(m) <= operand1(m)

    @_("operand BETWEEN operand AND operand")  # noqa: 821
    def condition(self, p):  # noqa: 811
        operand0 = p.operand0
        operand1 = p.operand1
        operand2 = p.operand2
        return lambda m: operand1(m) <= operand0(m) <= operand2(m)

    @_('operand IN "(" in_list ")"')  # noqa: 821
    def condition(self, p):  # noqa: 811
        operand = p.operand
        in_list = p.in_list
        return lambda m: operand(m) in in_list(m)

    @_("function")  # noqa: 821
    def condition(self, p):  # noqa: 811
        function = p.function
        return lambda m: function(m)

    @_("condition AND condition")  # noqa: 821
    def condition(self, p):  # noqa: 811
        condition0 = p.condition0
        condition1 = p.condition1
        return lambda m: condition0(m) and condition1(m)

    @_("condition OR condition")  # noqa: 821
    def condition(self, p):  # noqa: 811
        condition0 = p.condition0
        condition1 = p.condition1
        return lambda m: condition0(m) or condition1(m)

    @_("NOT condition")  # noqa: 821
    def condition(self, p):  # noqa: 811
        condition = p.condition
        return lambda m: not condition(m)

    @_("NOT path")  # noqa: 821
    def condition(self, p):  # noqa: 811
        return lambda m: not p.path(m)

    @_('"(" condition ")"')  # noqa: 821
    def condition(self, p):  # noqa: 811
        condition = p.condition
        return lambda m: condition(m)

    @_('ATTRIBUTE_EXISTS "(" path ")"')  # noqa: 821
    def function(self, p):  # noqa: 811
        path = p.path
        return lambda m: path(m) is not None

    @_('ATTRIBUTE_NOT_EXISTS "(" path ")"')  # noqa: 821
    def function(self, p):  # noqa: 811
        path = p.path
        return lambda m: path(m) is None

    @_('ATTRIBUTE_TYPE "(" path "," operand ")"')  # noqa: 821
    def function(self, p):  # noqa: 811
        path = p.path
        operand = p.operand
        return lambda x: path, operand

    @_('BEGINS_WITH "(" path "," operand ")"')  # noqa: 821
    def function(self, p):  # noqa: 811
        path = p.path
        operand = p.operand
        return (
            lambda m: path(m).startswith(operand(m))
            if isinstance(path(m), str)
            else False
        )

    @_('CONTAINS "(" path "," operand ")"')  # noqa: 821
    def function(self, p):  # noqa: 811
        path = p.path
        operand = p.operand
        return (
            lambda m: operand(m) in path(m)
            if isinstance(path(m), (str, set))
            else False
        )

    @_('SIZE "(" path ")"')  # noqa: 821
    def operand(self, p):  # noqa: 811
        path = p.path
        return (
            lambda m: len(path(m))
            if isinstance(path(m), (str, set, dict, bytearray, bytes, list))
            else -1
        )

    @_('in_list "," operand')  # noqa: 821
    def in_list(self, p):  # noqa: 811
        in_list = p.in_list
        operand = p.operand
        return lambda m: [*in_list(m), operand(m)]

    @_('operand "," operand')  # noqa: 821
    def in_list(self, p):  # noqa: 811
        operand0 = p.operand0
        operand1 = p.operand1
        return lambda m: [operand0(m), operand1(m)]

    @_("path")  # noqa: 821
    def operand(self, p):  # noqa: 811
        return p.path

    @_("VALUE")  # noqa: 821
    def operand(self, p):  # noqa: 811
        VALUE = p.VALUE
        if VALUE.startswith("'"):
            VALUE = VALUE.replace(r"\'", "'")
        else:
            VALUE = VALUE.replace(r"\"", '"')

        return lambda m: self.__strip_quotes(VALUE)

    @_("operand MATCH operand")  # noqa: 821
    def condition(self, f):  # noqa: 811
        regex = f.operand1
        str_to_match = f.operand0
        return lambda x: bool(re.match(regex(x), str_to_match(x)))

    @_('path "." NAME')  # noqa: 821
    def path(self, p):  # noqa: 811
        path = p.path
        NAME = p.NAME
        return lambda m: path(m).get(NAME) if isinstance(path(m), dict) else None

    @_('path "[" VALUE "]"')  # noqa: 821
    def path(self, p):  # noqa: 811
        key = self.__strip_quotes(p.VALUE)
        path = p.path

        return lambda m: path(m).get(key) if isinstance(path(m), dict) else None

    @_('path "[" INT "]"')  # noqa: 821
    def path(self, p):  # noqa: 811
        path = p.path
        INT = int(p.INT)
        return (
            lambda m: path(m)[INT]
            if isinstance(path(m), list) and len(path(m)) >= INT
            else None
        )

    @_("INT")  # noqa: 821
    def operand(self, o):  # noqa: 811
        INT = int(o.INT)
        return lambda m: INT

    @_("TRUE")  # noqa: 821
    def operand(self, o):  # noqa: 811
        return lambda m: True

    @_("FALSE")  # noqa: 821
    def operand(self, o):  # noqa: 811
        return lambda m: False

    @_("FLOAT")  # noqa: 821
    def path(self, p):  # noqa: 811
        FLOAT = float(p.FLOAT)
        return lambda m: FLOAT

    @_("KEYS")  # noqa: 821
    def path(self, _):  # noqa: 811
        return lambda m: m.keys

    @_("NEW_IMAGE")  # noqa: 821
    def path(self, _):  # noqa: 811
        return lambda m: m.new_image

    @_("OLD_IMAGE")  # noqa: 821
    def path(self, _):  # noqa: 811
        return lambda m: m.old_image

    @_("NAME")  # noqa: 821
    def path(self, p):  # noqa: 811
        NAME = p.NAME
        if isinstance(p, YaccProduction):
            raise KeywordError(f"Unknown keyword {p.NAME}")
        return lambda m: m.get(NAME) if p(m) is not None else None

    @_('FROM_JSON "(" path ")" ')  # noqa: 821
    def function(self, p):  # noqa: 811
        return lambda m: json.loads(p.path(m))

    @_('CHANGED "(" in_list ")"')  # noqa: 821
    @_('CHANGED "(" VALUE ")"')  # noqa: 821
    def function(self, p):  # noqa: 811
        # 1. Key is not in both dicts
        # 2. Key is in one and not the other
        # 3. Key is in both but the items differ
        if hasattr(p, "in_list"):
            key_list = p.in_list(p)
        else:
            key_list = [self.__strip_quotes(p.VALUE)]

        def has_changed(record, keys=key_list):
            for k in keys:
                if (
                    k not in record.old_image
                    and k in self.new_image
                    or k not in record.old_image
                    and k in record.new_image
                    or k in record.new_image
                    and k in record.old_image
                    and record.old_image[k] != record.new_image[k]
                ):
                    return True

            return False

        return has_changed

    @_('IS_TYPE "(" path "," NAME ")"')  # noqa: 821
    def function(self, p):  # noqa: 811
        path = p.path

        TYPE_MAP = {
            "B": lambda m: isinstance(path(m), Binary),
            "BOOL": lambda m: isinstance(path(m), bool),
            "BS": lambda m: isinstance(path(m), set)
            and [x for x in path(m) if isinstance(x, bytes)],
            "L": lambda m: isinstance(path(m), list),
            "M": lambda m: isinstance(path(m), dict),
            "N": lambda m: isinstance(path(m), Decimal),
            "NS": lambda m: isinstance(path(m), set)
            and [x for x in path(m) if isinstance(x, Decimal)],
            "NULL": lambda m: path(m) is None,
            "S": lambda m: isinstance(path(m), str),
            "SS": lambda m: isinstance(path(m), set)
            and [x for x in path(m) if isinstance(x, str)],
        }

        if p.NAME not in TYPE_MAP:
            raise TypeError(f"Unknown type '{p.NAME}'")

        return TYPE_MAP[p.NAME]

    @_("condition error")  # noqa: 821
    def operand(self, x):  # noqa: 811
        raise SyntaxError(f"Unexpected token '{x.error.value}'")



# pyright: reportUndefinedVariable=false
from sly import Lexer

from ..exceptions import SyntaxError


class ExpressionLexer(Lexer):
    # Set of token names.
    tokens = {
        VALUE,
        INT,
        FLOAT,
        BETWEEN,
        AND,
        OR,
        NOT,
        IN,
        EQ,
        NE,
        GT,
        GTE,
        LT,
        LTE,
        ATTRIBUTE_EXISTS,
        ATTRIBUTE_NOT_EXISTS,
        ATTRIBUTE_TYPE,
        BEGINS_WITH,
        CONTAINS,
        SIZE,
        FROM_JSON,
        KEYS,
        OLD_IMAGE,
        NEW_IMAGE,
        NAME,
        CHANGED,
        MATCH,
        IS_TYPE,
        FALSE,
        TRUE,
    }

    # Set of literal characters
    literals = {"(", ")", "[", "]", ",", "."}

    # String containing ignored characters
    ignore = " \t"

    # Regular expression rules for tokens
    OLD_IMAGE = r"\$OLD"
    NEW_IMAGE = r"\$NEW"

    AND = r"\&{1}"
    OR = r"\|{1}"
    NOT = "NOT"
    IN = "IN"
    BETWEEN = "BETWEEN"
    CHANGED = "has_changed"
    IS_TYPE = "is_type"
    ATTRIBUTE_EXISTS = "attribute_exists"
    ATTRIBUTE_NOT_EXISTS = "attribute_not_exists"
    ATTRIBUTE_TYPE = "attribute_type"
    BEGINS_WITH = "begins_with"
    CONTAINS = "contains"
    FROM_JSON = "from_json"
    SIZE = "size"
    TRUE = "True"
    FALSE = ("False",)

    """
    NAME has to come AFTER any keywords above. NAME is used as a path within OLD_IMAGE/NEW_IMAGE
    and also Dynamodb types such as S, L, SS, NS, BOOL, etc...
    """
    NAME = r"[a-zA-Z_][a-zA-Z0-9\-_]*"
    NE = "!="
    GTE = ">="
    LTE = "<="
    EQ = "=="
    GT = ">"
    LT = "<"
    INT = "\d+"
    MATCH = "=~"
    FLOAT = "\d+\.\d+"
    VALUE = r""""([^"\\]*(\\.[^"\\]*)*)"|\'([^\'\\]*(\\.[^\'\\]*)*)\'"""

    # Line number tracking
    @_(r"\n+")
    def ignore_newline(self, t):
        self.lineno += t.value.count("\n")

    def error(self, t):
        if t.value[0] == "$":
            raise SyntaxError(
                f"Invalid base path {t.value.split(' ')[0].split('.')[0]}"
            )
        else:
            raise SyntaxError(
                f"Bad character '{t.value[0]}' at line {self.lineno} character {self.index}"
            )



from __future__ import annotations

from concurrent.futures import Executor
from enum import Enum, auto
from functools import partial
from itertools import groupby
from types import FunctionType
from typing import Any, Callable

from multipledispatch import dispatch

from .conditions import Condition
from .conditions.parser import ExpressionParser
from .exceptions import RouteAlreadyExistsException
from .record import Record, RouteRecord


class Operation(Enum):
    INSERT = auto()
    MODIFY = auto()
    REMOVE = auto()


DYNAMODB_STREAM_ROUTER_NAMESPACE = dict()
dispatch = partial(dispatch, namespace=DYNAMODB_STREAM_ROUTER_NAMESPACE)

__CONDITION_PARSER = ExpressionParser()
__ROUTES: dict[Operation, set[Route]] = {operation: set() for operation in Operation}

RouteHandler = Callable[[RouteRecord], Any]


class Route:
    def __call__(self, record: RouteRecord) -> Any:
        return self.__handler(record)

    def __hash__(self) -> int:
        return hash(self.__condition)

    def __init__(
        self, *, condition: Condition, handler: RouteHandler, priority: int
    ) -> None:
        super().__init__()
        self.__condition = condition
        self.__priority = priority
        self.__handler = handler

    def __str__(self) -> str:
        return f"{self.__condition}[{self.__priority}] -> {self.__handler.__module__}.{self.__handler.__name__}"

    @property
    def _condition(self) -> Condition:
        return self.__condition

    @property
    def _handler(self) -> RouteHandler:
        return self.__handler

    @property
    def _priority(self) -> int:
        return self.__priority

    def match(self, record: RouteRecord) -> bool:
        return self.__condition(record)


def add_route(operation: Operation, route: Route) -> Route:
    if route:
        if has_route(operation, route):
            raise RouteAlreadyExistsException()
        __ROUTES[operation].add(route)
    return route


def get_routes(operation: Operation) -> frozenset[Route]:
    return frozenset(__ROUTES[operation])


def has_route(operation: Operation, route: Route) -> bool:
    return route in __ROUTES[operation]


def remove_route(operation: Operation, route: Route) -> Route:
    try:
        __ROUTES[operation].remove(route)
    except KeyError:
        pass
    return route


def update_route(operation: Operation, route: Route) -> Route:
    if route:
        __ROUTES[operation].add(route)
    return route


RouteHandlerDecorator = Callable[[RouteHandler], Route]


@dispatch(int)
def on_insert(priority: int, /) -> RouteHandlerDecorator:
    return on_insert(lambda _: True, priority)


@dispatch(str, int)
def on_insert(condition: str, priority: int, /) -> RouteHandlerDecorator:
    return on_operations({Operation.INSERT}, condition, priority)


@dispatch(FunctionType, int)
def on_insert(condition: Condition, priority: int, /) -> RouteHandlerDecorator:
    return on_operations({Operation.INSERT}, condition, priority)


@dispatch(int)
def on_modify(priority: int, /) -> RouteHandlerDecorator:
    return on_modify(lambda _: True, priority)


@dispatch(str, int)
def on_modify(condition: str, priority: int, /) -> RouteHandlerDecorator:
    return on_operations({Operation.MODIFY}, condition, priority)


@dispatch(FunctionType, int)
def on_modify(condition: Condition, priority: int, /) -> RouteHandlerDecorator:
    return on_operations({Operation.MODIFY}, condition, priority)


@dispatch(int)
def on_remove(priority: int, /) -> RouteHandlerDecorator:
    return on_remove(lambda _: True, priority)


@dispatch(str, int)
def on_remove(condition: str, priority: int, /) -> RouteHandlerDecorator:
    return on_operations({Operation.REMOVE}, condition, priority)


@dispatch(FunctionType, int)
def on_remove(condition: Condition, priority: int, /) -> RouteHandlerDecorator:
    return on_operations({Operation.REMOVE}, condition, priority)


@dispatch(set, int)
def on_operations(
    operations: set[Operation], priority: int, /
) -> RouteHandlerDecorator:
    return on_operations(operations, lambda _: True, priority)


@dispatch(set, str, int)
def on_operations(
    operations: set[Operation], condition: str, priority: int, /
) -> RouteHandlerDecorator:
    return on_operations(operations, __CONDITION_PARSER.parse(condition), priority)


@dispatch(set, FunctionType, int)
def on_operations(
    operations: set[Operation], condition: Condition, priority: int, /
) -> RouteHandlerDecorator:
    def register_route(handler: RouteHandler) -> RouteHandler:
        route = Route(condition=condition, handler=handler, priority=priority or 0)
        for operation in operations:
            add_route(operation, route)
        return handler

    return register_route


def route_record(
    record: Record,
    executor: Executor = None,
) -> None:
    operation = Operation[record["eventName"]]
    record: RouteRecord = RouteRecord(record)
    for _, routes in groupby(
        sorted(
            [route for route in __ROUTES[operation] if route.match(record)],
            key=lambda x: x._priority,
        ),
        key=lambda x: x._priority,
    ):
        routes = list(routes)
        routes[0](record) if len(routes) == 1 else list(
            (executor.map if executor else map)(
                Route.__call__, routes, [record] * len(routes)
            )
        )


def route_records(
    records: list[Record],
    executor: Executor = None,
) -> None:
    list(map(route_record, records, [executor] * len(records)))



from __future__ import annotations

from base64 import b64decode
from copy import deepcopy
from typing import Any, TypedDict

from boto3.dynamodb.types import TypeDeserializer

AttributeValueMap = dict[str, dict[str, Any]]


class Identity(TypedDict, total=False):
    PrincipalId: str
    Type: str


class StreamRecord(TypedDict, total=False):
    ApproximateCreationDateTime: int
    Keys: AttributeValueMap
    NewImage: AttributeValueMap
    OldImage: AttributeValueMap
    SequenceNumber: str
    SizeBytes: int
    StreamViewType: str


class Record(TypedDict, total=False):
    awsRegion: str
    dynamodb: StreamRecord
    eventID: str
    eventName: str
    eventSource: str
    eventSourceARN: str
    eventVersion: str
    userIdentity: Identity


class StreamTypeDeserializer(TypeDeserializer):
    def _deserialize_b(self, value):
        if isinstance(value, str):
            value = b64decode(value)
        return super()._deserialize_b(value)


class RouteRecord:
    __DESERIALIZER = StreamTypeDeserializer()

    def __init__(self, record: Record) -> None:
        self.__keys: dict[str, Any] = None
        self.__new_image: dict[str, Any] = None
        self.__old_image: dict[str, Any] = None
        self.__record = record

    @property
    def keys(self) -> dict[str, Any]:
        if self.__keys is None and "Keys" in self.__record["dynamodb"]:
            self.__keys = self.__DESERIALIZER.deserialize(
                dict(M=self.__record["dynamodb"]["Keys"])
            )
        return deepcopy(self.__keys)

    @property
    def new_image(self) -> dict[str, Any]:
        if self.__new_image is None and "NewImage" in self.__record["dynamodb"]:
            self.__new_image = self.__DESERIALIZER.deserialize(
                dict(M=self.__record["dynamodb"]["NewImage"])
            )
        return deepcopy(self.__new_image)

    @property
    def old_image(self) -> dict[str, Any]:
        if self.__old_image is None and "OldImage" in self.__record["dynamodb"]:
            self.__old_image = self.__DESERIALIZER.deserialize(
                dict(M=self.__record["dynamodb"]["OldImage"])
            )
        return deepcopy(self.__old_image)

    @property
    def record(self) -> Record:
        return deepcopy(self.__record)
